use super::*;

use winapi::um::errhandlingapi::{GetLastError, SetLastError};

/// Newtype for an error code value.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct ErrorCode(pub u32);
impl core::fmt::Display for ErrorCode {
  /// Attempts to format the `ErrorCode` as a human readable message obtained
  /// from Windows.
  ///
  /// If Windows can't give us a message, we fall back to giving the plan
  /// `Debug` format instead.
  #[allow(clippy::single_match_else)]
  fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
    use winapi::um::winbase::{
      FormatMessageW, FORMAT_MESSAGE_ALLOCATE_BUFFER, FORMAT_MESSAGE_FROM_SYSTEM,
      FORMAT_MESSAGE_IGNORE_INSERTS,
    };
    let dw_flags =
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;
    let lp_source = null_mut();
    let dw_message_id = self.0;
    let dw_language_id = 0;
    let mut buffer: *mut u16 = null_mut();
    let n_size = 0;
    let arguments = null_mut();
    let t_chars_stored_excluding_null = unsafe {
      FormatMessageW(
        dw_flags,
        lp_source,
        dw_message_id,
        dw_language_id,
        &mut buffer as *mut _ as *mut _,
        n_size,
        arguments,
      )
    };
    if buffer.is_null() {
      write!(f, "{:?}", self)
    } else {
      // TODO(Lokathor): We should develop a wrapper for the Local pointer so
      // that it can Drop properly so that we can use `?` in this messy stuff
      // without having to do a manual cleanup.
      let output = if t_chars_stored_excluding_null == 0 {
        write!(f, "{:?}", self)
      } else {
        let t_char_slice: &[u16] =
          unsafe { core::slice::from_raw_parts(buffer, t_chars_stored_excluding_null as usize) };
        let utf16_iter = core::char::decode_utf16(t_char_slice.iter().copied());
        let mut good = true;
        'write_loop: for decode_result in utf16_iter {
          match decode_result {
            Ok('\r') | Ok('\n') => continue,
            Ok(ch) => match write!(f, "{}", ch) {
              Ok(()) => continue,
              Err(_) => {
                good = false;
                break 'write_loop;
              }
            },
            Err(_) => {
              good = false;
              break 'write_loop;
            }
          }
        }
        if good {
          Ok(())
        } else {
          Err(core::fmt::Error)
        }
      };
      unsafe { local_free(buffer as *mut _).map_err(|_| core::fmt::Error)? };
      output
    }
  }
}
impl ErrorCode {
  /// Bit that signifies if an error code is an application error.
  pub const APPLICATION_ERROR_BIT: u32 = 1 << 29;

  /// Checks if this error code is an application error code.
  #[inline(always)]
  pub const fn is_application_error(self) -> bool {
    (self.0 & Self::APPLICATION_ERROR_BIT) > 0
  }
}

/// The calling thread's last-error code.
///
/// See [`GetLastError`](https://docs.microsoft.com/en-us/windows/win32/api/errhandlingapi/nf-errhandlingapi-getlasterror)
#[inline(always)]
pub fn get_last_error() -> ErrorCode {
  ErrorCode(unsafe { GetLastError() })
}

/// Sets the calling thread's last-error code.
///
/// See [`SetLastError`](https://docs.microsoft.com/en-us/windows/win32/api/errhandlingapi/nf-errhandlingapi-setlasterror)
#[inline(always)]
pub fn set_last_error(error: ErrorCode) {
  unsafe { SetLastError(error.0) }
}
