#![warn(missing_docs)]
#![cfg(windows)]
#![allow(clippy::if_not_else)]

//! Easier wrappers for Win32 API stuff, safe when possible.

use core::{
  convert::TryFrom,
  ffi::c_void,
  ptr::{null, null_mut, NonNull},
};
use winapi::{
  shared::{
    minwindef::{ATOM, HINSTANCE, HMODULE},
    windef::{HBRUSH, HCURSOR, HDC, HICON, HMENU, HWND, POINT, RECT},
  },
  um::{winnt::HANDLE, winuser::WNDCLASSW},
};

mod consoleapi;
pub use consoleapi::*;

mod errhandlingapi;
pub use errhandlingapi::*;

mod libloaderapi;
pub use libloaderapi::*;

mod processenv;
pub use processenv::*;

mod profileapi;
pub use profileapi::*;

mod winbase;
pub use winbase::*;

mod wingdi;
pub use wingdi::*;

mod winuser;
pub use winuser::*;

/// Allocates a utf-16, null-terminated version of the `&str` given.
pub fn wide_null(s: &str) -> Vec<u16> {
  s.encode_utf16().chain(Some(0)).collect()
}

/// Newtype of [`ATOM`](winapi::shared::minwindef::ATOM).
///
/// This helps you avoid accidentally mixing an `ATOM` with other `u16` values,
/// but otherwise has no additional semantics.
#[repr(transparent)]
#[derive(Debug, Clone, Copy, Default, PartialEq, Eq)]
pub struct Atom(pub ATOM);

/// Laid out like [`POINT`](https://docs.microsoft.com/en-us/windows/win32/api/windef/ns-windef-point)
#[repr(C)]
#[derive(Debug, Clone, Copy, Default, PartialEq, Eq)]
#[allow(missing_docs)]
pub struct Point {
  pub x: i32,
  pub y: i32,
}

/// Rect that stores a corner point as well as a width and height.
#[repr(C)]
#[derive(Debug, Clone, Copy, Default, PartialEq, Eq)]
#[allow(missing_docs)]
pub struct PointSizeRect {
  pub x: i32,
  pub y: i32,
  pub width: i32,
  pub height: i32,
}

/// Rect that stores four edge locations.
///
/// Laid out like [`RECT`](https://docs.microsoft.com/en-us/windows/win32/api/windef/ns-windef-rect)
#[repr(C)]
#[derive(Debug, Clone, Copy, Default, PartialEq, Eq)]
#[allow(missing_docs)]
pub struct EdgeRect {
  pub left: i32,
  pub top: i32,
  pub right: i32,
  pub bottom: i32,
}
