use super::*;

use winapi::um::winbase::LocalFree;

/// The the standard handles of a process.
///
/// For use with [`get_std_handle`](get_std_handle) and
/// [`set_std_handle`](set_std_handle)
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u32)]
pub enum StdHandle {
  /// The standard input device.
  ///
  /// Initially, this is the console input buffer, `CONIN$`.
  Input = winapi::um::winbase::STD_INPUT_HANDLE,
  /// The standard output device.
  ///
  /// Initially, this is the active console screen buffer, `CONOUT$`.
  Output = winapi::um::winbase::STD_OUTPUT_HANDLE,
  /// The standard error device.
  ///
  /// Initially, this is the active console screen buffer, `CONOUT$`.
  Error = winapi::um::winbase::STD_ERROR_HANDLE,
}

/// Frees/invalidates a Local memory handle.
///
/// ## Safety
///
/// * The pointer must have been allocated by
///   [`LocalAlloc`](https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-localalloc)
///   or
///   [`LocalReAlloc`](https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-localrealloc)
///
/// See
/// [`LocalFree`](https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-localfree)
#[inline]
pub unsafe fn local_free(mem: *mut c_void) -> Result<(), ErrorCode> {
  // TODO: in 0.4 winapi will use the same `c_void` as core so we won't need to
  // pointer cast here.
  let output = LocalFree(mem as *mut _);
  if output.is_null() {
    Ok(())
  } else {
    Err(get_last_error())
  }
}
